# Duo, A crazy eights game.
# Copyright (C) 2003-2008 The Duo developers.
#
# This file is part of Duo.
#
# Duo is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

"""
TODO
"""

import cPickle
import logging
import threading

from duo.transport import Context
from duo.transport import Endpoint
from duo.transport import TransportJoinPacket
from duo.transport import TransportEscapePacket


# The module must be imported after the logger was configured.
logger = logging.getLogger('duo.transport.local')


# The queue contains tuples of the form (endpoint, payload, from)

# TODO: context destruction: check every time an endpoint quit a context.
#       contexts should be destroyed if empty.

# Singleton
class LocalTransportLayer(threading.Thread):
    """TODO
    """
    def __new__(cls, *p, **k):
        """TODO"""
        if not '_instance' in cls.__dict__:
            cls._instance = object.__new__(cls)
            threading.Thread.__init__(cls._instance, name="LocalTransportLayer")
            cls._instance.__dict__['_starting'] = True
            cls._instance.__dict__['_root_context'] = Context('root')
            cls._instance.__dict__['_endpoints'] = []
            cls._instance.__dict__['_lock'] = threading.Lock()
            cls._instance.__dict__['_queue'] = []
            cls._instance.__dict__['_semaphore'] = threading.Semaphore(0)
            cls._instance.start()
            logger.debug('LocalTransportLayer created.')
        return cls._instance

    def __init__(self):
        pass

    def run(self):
        """TODO"""
        logger.debug('LocalTransportLayer starting.')
        while len(self._endpoints) or len(self._queue) or self._starting:
            self._starting = False
            self._semaphore.acquire()
            self._lock.acquire()
            try:
                if len(self._queue):
                    endpoint, payload, emitter = self._queue.pop(0)
                else:
                    continue
            finally:
                self._lock.release()

            #logger.debug('sending payload of size %d from %s to %s'
            #             % (len(payload), emitter, endpoint))
            if emitter:
                emitter = emitter.get_id()
            else:
                emitter = None
            endpoint.notify(payload, emitter)
        logger.debug('LocalTransportLayer leaving.')

    def attach(self, endpoint, name='root'):
        """Connect to a context."""
        self._lock.acquire()
        try:
            # Store the endpoint
            if not endpoint in self._endpoints:
                self._endpoints.append(endpoint)

            # Join the new context. Eventually create it.
            context = self._root_context.get_or_create_context(name)
            self._raise_packet(
                TransportJoinPacket(endpoint.get_id()), context)
            context.add_endpoint(endpoint)
        finally:
            self._lock.release()

    def detach(self, endpoint):
        """Leave a context."""
        self._lock.acquire()
        try:
            # Check if endpoint is connected
            context = endpoint.get_context()
            if not context:
                return

            context.remove_endpoint(endpoint)
            if len(context.get_endpoints()):
                self._raise_packet(
                    TransportEscapePacket(endpoint.get_id()), context)
            else:
                # If there are no mode endpoints in the context, the
                # loop must still be awaken once to check if it should
                # exit.
                self._semaphore.release()

            # Check if endpoint is to be disconnected
            if context == self._root_context:
                self._endpoints.remove(endpoint)

            # TODO: check if context should be deleted
        finally:
            self._lock.release()

    def get_contexts(self):
        """TODO

        can raise BadContext is the requester is not attached to root context"""
        contexts = []
        self._lock.acquire()
        try:
            contexts = [context.get_name() for context
                        in self._root_context.get_child_contexts()]
        finally:
            self._lock.release()
        return contexts

    def send_to(self, payload, emitter, to):
        """Send a message to a endpoint."""
        self._lock.acquire()
        try:
            for endpoint in emitter.get_context().get_endpoints():
                if endpoint.get_id() == to:
                    self._queue.append((endpoint, payload, emitter))
                    self._semaphore.release()
                    break
        finally:
            self._lock.release()

    def broadcast(self, payload, emitter):
        """Send a message to every connected endpoint."""
        self._lock.acquire()
        try:
            for endpoint in emitter.get_context().get_endpoints():
                if endpoint != emitter:
                    self._queue.append((endpoint, payload, emitter))
                    self._semaphore.release()
        finally:
            self._lock.release()

    def _raise_packet(self, packet, context):
        # Warning. Lock must be managed by the caller.
        data = cPickle.dumps(packet)
        for endpoint in context.get_endpoints():
            self._queue.append((endpoint, data, None))
            self._semaphore.release()


class LocalEndpoint(Endpoint):
    """TODO
    """
    def __init__(self, id, options):
        Endpoint.__init__(self, id, options['callbacks'])
        self._transport = LocalTransportLayer()
        self._transport.attach(self, options['context'])

    def get_contexts(self):
        """Return full information on existing contexts."""
        return self._transport.get_contexts()

    def join_context(self, name):
        """Attach this endpoint to the given context."""
        self._transport.attach(self, name)

    def escape(self):
        """Detach from the current context.

        Once detached, the endpoint is automatically attached to the
        parent context. If there is no parent, the endpoint is disconnected.
        """
        self._transport.detach(self)

    def _do_send(self, packet, to):
        """Send a packet.

        If aTo is None, the packet has to be broadcasted.
        """
        if to is None:
            self._transport.broadcast(packet, self)
        else:
            self._transport.send_to(packet, self, to)


OPTIONS = []
CLASS   = LocalEndpoint
